/* format command
:ts=4

   TODO

   - Retry.
   - GUI.
   - Implement other missing stuff. See #warning's.
*/

#define DEBUG_FMTPART(x)			;


#include <exec/memory.h>
#include <exec/io.h>
#include <exec/errors.h>
#include <devices/trackdisk.h>
#include <devices/scsidisk.h>
#include <dos/dos.h>
#include <dos/dosextens.h>
#include <dos/dosasl.h>
#include <dos/exall.h>
#include <dos/filehandler.h>
#include <utility/tagitem.h>
#include <workbench/icon.h>
#include <workbench/workbench.h>

#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/utility.h>
#include <proto/locale.h>
#include <proto/icon.h>

#include "Format_cat.h"
#include "format_version.h"

static const char version[] = VERSTAG;


#define SFS_PACKET_BASE         (0xf00000)

#define SFS_INQUIRY             (SFS_PACKET_BASE + 0x11000)

#define ACTION_GET_TOTAL_BLOCKS (SFS_INQUIRY + 0)
#define ACTION_GET_BLOCKSIZE    (SFS_INQUIRY + 1)
#define ACTION_GET_BLOCKDATA    (SFS_INQUIRY + 2)

#define ACTION_SET_DEBUG        (SFS_PACKET_BASE + 0xBDC0)
#define ACTION_SET_CACHE        (SFS_PACKET_BASE + 0xBDC0 + 1)
#define ACTION_FORMAT_ARGS      (SFS_PACKET_BASE + 0xBDC0 + 2)

/* Above are 'old' packet types.  Below are the new types. */

#define ACTION_SFS_QUERY             (SFS_PACKET_BASE + 1)
#define ACTION_SFS_SET               (SFS_PACKET_BASE + 2)

#define ACTION_SFS_SET_OBJECTBITS    (SFS_PACKET_BASE + 100 + 1)
#define ACTION_SFS_LOCATE_OBJECT     (SFS_PACKET_BASE + 100 + 2)
#define ACTION_SFS_FORMAT            (SFS_PACKET_BASE + 100 + 3)

#define ACTION_SFS_READ_BITMAP       (SFS_PACKET_BASE + 200 + 1)

#define ACTION_SFS_DEFRAGMENT_INIT     (SFS_PACKET_BASE + 300 + 1)
#define ACTION_SFS_DEFRAGMENT_STEP     (SFS_PACKET_BASE + 300 + 2)

/* Tags used by ACTION_SFS_FORMAT: */

#define ASFBASE            (TAG_USER)

#define ASF_NAME           (ASFBASE + 1)  /* The name of the disk.  If this tag is not specified
format will fail with ERROR_INVALID_COMPONENT_NAME. */

#define ASF_NORECYCLED     (ASFBASE + 2)  /* If TRUE, then this tag will prevent the Recycled
directory from being created.  It defaults to creating
the Recycled directory. */

#define ASF_CASESENSITIVE  (ASFBASE + 3)  /* If TRUE, then this tag will use case sensitive file
and directory names.  It defaults to case insensitive
	names. */

#define ASF_SHOWRECYCLED   (ASFBASE + 4)  /* If TRUE, then the Recycled directory will be visible
	in directory listings.  It defaults to an invisible
	Recycled directory. */

#define ASF_RECYCLEDNAME   (ASFBASE + 5)  /* The name of the Recycled directory.  Defaults to
	'.recycled'. */


#ifndef CMD_READ64
#define CMD_READ64   24
#define CMD_WRITE64  25
#define CMD_SEEK64   26
#define CMD_FORMAT64 27
#endif
#define io_HighOffset io_Actual
#define io_LowOffset  io_Offset


static
LONG DoFormat(void);

static
BOOL OpenDev(void);

static
void CloseDev(void);

static
LONG QuickFormat(ULONG dostype);

static
LONG WaitValidated(ULONG dostype);

static
BOOL FormatPartition(ULONG dostype);

static
CONST_STRPTR ConvDevErr(LONG err, LONG *ioerr);

static
ULONG CreateIcons(CONST_STRPTR Volume);


static const
UBYTE template[]=	"DEVICE=DRIVE/A/K,"
"NAME/A/K,"
"OFS/S,"
"FFS/S,"
"SFS/S,"
"MSDOS/S,"
"INTL=INTERNATIONAL=CASESENSITIVE/S,"
"NOINTL=NOINTERNATIONAL/S,"
"DIRCACHE/S,"
"NODIRCACHE/S,"
"LNFS=LONGFILENAMES/S,"
"NOLNFS=NOLONGFILENAMES/S,"
"NOICONS/S,"
"QUICK/S,"
"NORECYCLED/S,"
"SHOWRECYCLED/S";

struct
{
    char *device;
    char *name;
    ULONG ofs;
    ULONG ffs;
    ULONG sfs;
    ULONG msdos;
    ULONG international;
    ULONG nointernational;
    ULONG dircache;
    ULONG nodircache;
    ULONG lnfs;
    ULONG nolnfs;
    ULONG noicons;
    ULONG quick;
    ULONG norecycled;
    ULONG showrecycled;
}
arglist;

struct ExecBase		*SysBase;
struct DosLibrary	*DOSBase;
struct Library		*UtilityBase;
struct Library		*LocaleBase;

struct MsgPort *FSPort;
struct MsgPort IOMsgPort;
struct IOStdReq *IOReq = NULL;
BPTR input, output;
struct DosEnvec DE;
UBYTE DevName[256];
ULONG DevUnit, DevFlags;
BOOL WritesDone = FALSE;
CONST_STRPTR errhead;

int main(void)
{
	LONG ret = RETURN_FAIL;
	struct RDArgs *rdargs;

	IOMsgPort.mp_SigBit = -1;

	SysBase = *((struct ExecBase**) 4);
	DOSBase = (struct DosLibrary *)OpenLibrary("dos.library",37);
	UtilityBase = OpenLibrary("utility.library",37);
	if (DOSBase && UtilityBase)
	{
		LocaleBase = OpenLibrary("locale.library", 37);
		OpenFormatCatalog();

		errhead = GetStr(MSG_TITLE_FORMAT_FAILURE);

		rdargs = ReadArgs((STRPTR) template, (LONG *)&arglist, NULL);
		if (rdargs)
		{
			static CONST_STRPTR banned[] =
			{
				"MOSSYS", "SYS", "L", "DEVS", "LIBS", "S", "C",
				/* "LOCALE", "ENVARC", "ENV", "KEYMAPS", "PRINTERS", */
				NULL
			};
			CONST_STRPTR *banneddevnames = banned;
			CONST_STRPTR *bannedvolnames = banned;
#ifdef __MORPHOS__
			if ((DOSBase->dl_lib.lib_Version == 50 && DOSBase->dl_lib.lib_Revision >= 50) ||
			    DOSBase->dl_lib.lib_Version > 50)
			{
				banneddevnames = DOSBase->dl_BannedDeviceNames;
				bannedvolnames = DOSBase->dl_BannedVolumeNames;
			}
#endif
			/* Check for offending names that potentially could hose the system
			 */
			for ( ; *banneddevnames; banneddevnames++)
			{
				if (!Stricmp(*banneddevnames, arglist.name))
				{
					SetIoErr(ERROR_INVALID_COMPONENT_NAME);

					break;
				}
			}
			if (!*banneddevnames)
			{
				for ( ; *bannedvolnames; bannedvolnames++)
				{
					if (!Stricmp(*bannedvolnames, arglist.name))
					{
						SetIoErr(ERROR_INVALID_COMPONENT_NAME);

						break;
					}
				}
			}

			if (!*banneddevnames && !*bannedvolnames)
			{
				struct DosList *dl;
				UBYTE *p;

				p = arglist.device;
				while (*p)
				{
					if (*p == ':')
					{
						*p = '\0';
						break;
					}
					p++;
				}

				ret = RETURN_ERROR;

				dl = LockDosList(LDF_DEVICES | LDF_READ);
				dl = FindDosEntry(dl, arglist.device, LDF_DEVICES);
				if (dl)
				{
					p = BADDR(dl->dol_Name);
					memcpy(arglist.device, p + 1, p[0]);
					arglist.device[p[0]] = '\0';

					FSPort = dl->dol_Task;

					UnLockDosList(LDF_DEVICES | LDF_READ);

					input = Input();
					output = Output();

					ret = DoFormat();
				}
				else
				{
					UnLockDosList(LDF_DEVICES | LDF_READ);
					PrintFault(ERROR_DEVICE_NOT_MOUNTED, errhead);
					SetIoErr(ERROR_DEVICE_NOT_MOUNTED);
				}
			}
			else
			{
				PrintFault(IoErr(), errhead);
			}

			FreeArgs(rdargs);
		}
		else
		{
			PrintFault(IoErr(), errhead);
		}

		CloseFormatCatalog();
		if (LocaleBase)
		{
			CloseLibrary(LocaleBase);
		}
	}
	if (UtilityBase)
	{
		CloseLibrary(UtilityBase);
	}
	if (DOSBase)
	{
		CloseLibrary((struct Library *)DOSBase);
	}

	return ret;
}


static
LONG DoFormat(void)
{
	LONG ret = RETURN_ERROR;
	ULONG Status = FALSE;

	Printf(GetStr(MSG_BODY_INSERT_DISK_TO_BE_FORMATTED), (LONG) arglist.device);
	Flush(output);

	if (!IsInteractive(input))
	{
		return FALSE;
	}

	for (;;)
	{
		if (WaitForChar(input, 100))
		{

#warning "TODO: Check if the disk is write protected etc..."

			if (DoPkt(FSPort, ACTION_INHIBIT, DOSTRUE, 0, 0, 0, 0))
			{
				BOOL Inhibited = TRUE;
				ULONG dostype = 0;

				if (OpenDev())
				{
					if (DE.de_TableSize >= DE_DOSTYPE)
					{
						dostype = DE.de_DosType;
					}

					if (!dostype)
					{
						if (arglist.sfs)
						{
							dostype = 0x53465300;
						}
						else
						{
							if (arglist.msdos)
							{
								dostype = ID_MSDOS_DISK;
							}
							else if (arglist.ffs)
							{
								dostype = ID_FFS_DISK;
							}
							else if (arglist.ofs)
							{
								dostype = ID_DOS_DISK;
							}
							else /* default */
							{
								dostype = ID_FFS_DISK;
							}
						}
					}

					if ((dostype & 0xffffff00) == ID_DOS_DISK)
					{
						if (arglist.ofs)
						{
							dostype = ID_DOS_DISK;
						}
						else
						{
							if (arglist.ffs)
							{
								dostype |= 1;
							}

							if (arglist.international &&
							    !arglist.nointernational)
							{
								dostype |= 2;
							}
							else if (arglist.nointernational &&
							         !arglist.international)
							{
								dostype &= ~2;
							}
							if (arglist.dircache &&
							    !arglist.nodircache)
							{
								dostype |= 4;
							}
							else if (arglist.nodircache &&
							         !arglist.dircache)
							{
								dostype &= ~4;
							}
							if (arglist.lnfs &&
							    !arglist.nolnfs)
							{
								dostype |= (2 | 4);
							}
							else if (arglist.nolnfs &&
							         !arglist.lnfs)
							{
								dostype &= ~4; /* Yes, just turn off DC! -Piru */
							}
						}
					}

					if (!arglist.quick)
					{
						PutStr(GetStr(MSG_FORMATTING_DISK));
						PutStr("\x9b" "0 p\n");

						Status = FormatPartition(dostype);

						PutStr("\x9b" " p"); Flush(output);
					}
					else
						Status = TRUE;

					if (Status)
					{
						PutStr("\n");
						PutStr(GetStr(MSG_INITIALIZING_DISK));
						Flush(output);

						if (arglist.sfs)
						{
							struct TagItem tags[5];
							struct TagItem *tag = tags;

							tag->ti_Tag  = ASF_NAME;
							tag->ti_Data = (ULONG)arglist.name;
							tag++;

							if (arglist.international)
							{
								tag->ti_Tag  = ASF_CASESENSITIVE;
								tag->ti_Data = TRUE;
								tag++;
							}

							if (arglist.norecycled)
							{
								tag->ti_Tag  = ASF_NORECYCLED;
								tag->ti_Data = TRUE;
								tag++;
							}

							if (arglist.showrecycled)
							{
								tag->ti_Tag  = ASF_SHOWRECYCLED;
								tag->ti_Data = TRUE;
								tag++;
							}

							tag->ti_Tag = TAG_END;

							if (!DoPkt(FSPort, ACTION_SFS_FORMAT, (LONG)&tags, 0, 0, 0, 0))
							{
								Status = FALSE;
								PrintFault(IoErr(), errhead);
							}
						}
						else
						{
							if (!QuickFormat(dostype))
							{
								Status = FALSE;
								PrintFault(IoErr(), errhead);
							}
						}

						DoPkt(FSPort, ACTION_INHIBIT, DOSFALSE, 0, 0, 0, 0);
						Inhibited = FALSE;

						if (Status)
						{
							PutStr("\n");

							Status = WaitValidated(dostype);

							if (Status)
							{
								if (!arglist.noicons)
								{
									CreateIcons(arglist.device);
								}

								ret = RETURN_OK;
							}
						}
					}
					else
					{
						PrintFault(IoErr(), errhead);
					}

					CloseDev();
				}
				else
				{
					PrintFault(IoErr(), errhead);
					Printf(GetStr(MSG_BODY_COULD_NOT_OPEN_DEVICE), (LONG) arglist.device);
					PutStr("\n");
				}

				if (Inhibited)
				{
					DoPkt(FSPort, ACTION_INHIBIT, DOSFALSE, 0, 0, 0, 0);
				}
			}
			else
			{
				PrintFault(IoErr(), errhead);
			}

			break;
		}
		else if (SetSignal(0L, SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C)
		{
			PutStr("\n");
			PrintFault(ERROR_BREAK, NULL);
			ret = RETURN_WARN;

			break;
		}
	}

	while (WaitForChar(input, 0))
	{
		FGetC(input);
	}

	return ret;
}


static
LONG QuickFormat(ULONG dostype)
{
	LONG ret = FALSE;
	UBYTE *name;
	ULONG len;

	len = strlen(arglist.name);
	if (len < 255 && (name = AllocMem(len + 1, MEMF_PUBLIC)))
	{
		LONG rc;

		name[0] = len;
		memcpy(name + 1, arglist.name, len);

		rc = DoPkt(FSPort, ACTION_FORMAT, MKBADDR(name), dostype, 0, 0, 0);

		FreeMem(name, len + 1);

		if (rc)
		{
			ret = TRUE;
		}
		else if (IoErr() == ERROR_ACTION_NOT_KNOWN)
		{
#warning "TODO: Manual FFS initialization..."
		}
	}
	else
		SetIoErr(ERROR_NO_FREE_STORE);

	return ret;
}


static
LONG WaitValidated(ULONG dostype)
{
	UBYTE _id[sizeof(struct InfoData) + 3];
	struct InfoData *id = (struct InfoData *) (((ULONG) _id + 3) & ~3);
	int cnt;

	for (cnt = 0; cnt < 100; cnt++)
	{
		if (SetSignal(0L, SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C)
		{
			SetIoErr(ERROR_BREAK);
			return FALSE;
		}

		Delay(15);

		if (DoPkt(FSPort, ACTION_DISK_INFO, MKBADDR(id), 0, 0, 0, 0))
		{
			if (id->id_DiskState == ID_VALIDATING)
			{
				continue;
			}

			if (dostype)
			{
				if (id->id_DiskType == dostype ||
				    (id->id_DiskType >= ID_DOS_DISK &&
				     id->id_DiskType <= ID_LNFS_FFS_DISK))
				{
					return TRUE;
				}

				if (cnt > 1)
				{
					break;
				}
			}
			else
			{
				return TRUE;
			}
		}
	}

	SetIoErr(ERROR_NOT_A_DOS_DISK);
	return FALSE;
}


static
BOOL OpenDev(void)
{
	BOOL doslistlocked = TRUE;
	struct DosList *dl;
	LONG ioerr = 0;

	dl = LockDosList(LDF_DEVICES | LDF_READ);
	dl = FindDosEntry(dl, arglist.device, LDF_DEVICES);
	if (dl)
	{
		struct FileSysStartupMsg *fssm;

		fssm = BADDR(dl->dol_misc.dol_handler.dol_Startup);

		if ((ULONG) fssm > 0x400 &&
	    	!(fssm->fssm_Unit & 0xff000000) &&
	    	fssm->fssm_Device && fssm->fssm_Environ)
		{
			struct DosEnvec *DosEnvec;

			DosEnvec = BADDR(fssm->fssm_Environ);
			if (DosEnvec->de_TableSize >= DE_UPPERCYL)
			{
				IOMsgPort.mp_SigBit = AllocSignal(-1);
				if (IOMsgPort.mp_SigBit != (UBYTE) -1)
				{
					IOMsgPort.mp_Node.ln_Type = NT_MSGPORT;
					IOMsgPort.mp_Flags   = PA_SIGNAL;
					IOMsgPort.mp_SigTask = FindTask(NULL);
					NEWLIST(&IOMsgPort.mp_MsgList);

					IOReq = CreateIORequest(&IOMsgPort, sizeof(*IOReq));
					if (IOReq)
					{
						ULONG len;
						UBYTE *str;
						LONG err;

						len = (DosEnvec->de_TableSize + 1) * sizeof(ULONG);
						if (len >= sizeof(struct DosEnvec))
						{
							len = sizeof(struct DosEnvec);
						}
						memcpy(&DE, DosEnvec, len);

						/* Set *SAFE* defaults if some fields are missing */
						if (DE.de_TableSize < DE_MASK)
						{
							/* longword alignment */
							DE.de_Mask = 0xfffffffc;
						}
						if (DE.de_TableSize < DE_MAXTRANSFER)
						{
							/* 64k - 512bytes */
							DE.de_MaxTransfer = 0x0000fe00;
						}
						if (DE.de_TableSize < DE_BUFMEMTYPE)
						{
							/* Public memory */
							DE.de_BufMemType = MEMF_PUBLIC /* | MEMF_24BITDMA */ ;
						}

						if (!DE.de_Mask)
						{
							DE.de_Mask = 0xfffffffc;
						}

						/* Make sure maxtransfer is sane */
						DE.de_MaxTransfer &= (ULONG) -(DE.de_SizeBlock * sizeof(ULONG));
						if (DE.de_MaxTransfer < (DE.de_SizeBlock * sizeof(ULONG)))
						{
							DE.de_MaxTransfer = DE.de_SizeBlock * sizeof(ULONG);
						}
#ifdef __MORPHOS__
						DE.de_BufMemType &= ~(MEMF_CHIP | MEMF_LOCAL | MEMF_24BITDMA | MEMF_CLEAR);
#endif
						DE.de_BufMemType |= MEMF_PUBLIC;

						str = BADDR(fssm->fssm_Device);
						memcpy(DevName, str + 1, str[0]); DevName[str[0]] = '\0';
						DevUnit = fssm->fssm_Unit;
						DevFlags = fssm->fssm_Flags;

						UnLockDosList(LDF_DEVICES | LDF_READ);
						doslistlocked = FALSE;

						err = OpenDevice(DevName, DevUnit, (APTR) IOReq, DevFlags);

						if (!err)
						{
							return TRUE;
						}

						ConvDevErr(err, &ioerr);

						DeleteIORequest((APTR) IOReq);
						IOReq = NULL;
					}
					else
						ioerr = ERROR_NO_FREE_STORE;

					FreeSignal(IOMsgPort.mp_SigBit);
					IOMsgPort.mp_SigBit = -1;
				}
			}
			else
				ioerr = ERROR_OBJECT_WRONG_TYPE;
		}
		else
			ioerr = ERROR_OBJECT_WRONG_TYPE;
	}
	else
		ioerr = ERROR_DEVICE_NOT_MOUNTED;

	if (doslistlocked)
	{
		UnLockDosList(LDF_DEVICES | LDF_READ);
	}

	SetIoErr(ioerr);

	return FALSE;
}


static
void CloseDev(void)
{
	if (IOReq)
	{
		/* Flush possible write buffers before stopping motor */

		if (WritesDone)
		{
			IOReq->io_Command = CMD_UPDATE;
			IOReq->io_Length  = 0;
			IOReq->io_Data    = NULL;
			if (DoIO((APTR) IOReq) != 0)
			{
				WritesDone = FALSE;
			}
		}
		if (WritesDone)
		{
			IOReq->io_Command = CMD_CLEAR;
			IOReq->io_Length  = 0;
			IOReq->io_Data    = NULL;
			DoIO((APTR) IOReq);
		}

		/* Stop motor */

		IOReq->io_Command = TD_MOTOR;
		IOReq->io_Length  = 0;
		IOReq->io_Data    = NULL;
		DoIO((APTR) IOReq);

		CloseDevice((APTR) IOReq);

		DeleteIORequest((APTR) IOReq);
		IOReq = NULL;
	}

	if (IOMsgPort.mp_SigBit != (UBYTE) -1)
	{
		FreeSignal(IOMsgPort.mp_SigBit);
		IOMsgPort.mp_SigBit = -1;
	}
}


static
void UpdateStats(BOOL printnow, LONG mode, ULONG curcyl, ULONG numcyls);

static
BOOL FormatPartition(ULONG dostype)
{
	BOOL ret = FALSE;
	LONG err, ioerr = 0;
	UBYTE *memptr;
	ULONG *readbuf;
	ULONG *writebuf;
	ULONG tracksize;
	ULONG startcyl, numcyls, curcyl;
	/* buffer alloc */
	BOOL usetd64 = FALSE;
	UQUAD offset;
	ULONG memlen, blocksize, align, bufsize;
	LONG i, mode;

	blocksize = DE.de_SizeBlock * sizeof(ULONG);
	tracksize = blocksize * DE.de_BlocksPerTrack * DE.de_Surfaces;

	/* Figure out the aligment requirement
	 */
	align = 0x01000000;
	for (i = 0; i < 24; i++)
	{
		if (DE.de_Mask & (1UL << i))
		{
			align = 1UL << i;
			break;
		}
	}
	DEBUG_FMTPART(Printf("blocksize: %lu tracksize: %lu align: %lu\n", blocksize, tracksize, align));

    /* Does the partition span to >4gig area ?
     */
	offset = (UQUAD) DE.de_HighCyl * (UQUAD) tracksize;
	if (offset > 4294967296ULL)
	{
		memlen = blocksize + align - 1;
		memptr = AllocMem(memlen, DE.de_BufMemType);
		if (!memptr)
		{
			return FALSE;
		}
		readbuf = (APTR) ((((ULONG) memptr) + align - 1) & (ULONG) -align);

		/* Check if TD64 works
		*/
		offset = (UQUAD) DE.de_HighCyl * (UQUAD) tracksize - blocksize;
		IOReq->io_Command    = CMD_READ64;
		IOReq->io_Flags      = 0;
		IOReq->io_Length     = blocksize;
		IOReq->io_Data       = readbuf;
		IOReq->io_HighOffset = offset >> 32;
		IOReq->io_LowOffset  = offset & 0xffffffff;

		err = DoIO((APTR) IOReq);

		FreeMem(memptr, memlen);

		if (err == IOERR_NOCMD)
		{
			SetIoErr(ERROR_NOT_IMPLEMENTED);
			return FALSE;
		}

		usetd64 = TRUE;
	}

	ret = TRUE;

	/* Limit bufsize by maxtransfer
     */
	bufsize  = (tracksize > DE.de_MaxTransfer) ? DE.de_MaxTransfer : tracksize;
	startcyl = DE.de_LowCyl;
	numcyls  = DE.de_HighCyl - DE.de_LowCyl;

	DEBUG_FMTPART(Printf("startcyl: %lu numcyls: %lu\n", startcyl, numcyls));

	memlen = (bufsize + align) * 2;
	memptr = AllocMem(memlen, DE.de_BufMemType);
	if (!memptr)
	{
		return FALSE;
	}
	writebuf = (APTR) ((((ULONG) memptr) + align - 1) & (ULONG) -align);
	readbuf  = (APTR) ((((ULONG) writebuf) + bufsize + align - 1) & (ULONG) -align);

	for (i = 0; i < (bufsize / sizeof(ULONG)); i++)
	{
		writebuf[i] = (dostype & 0xffffff00) | (i & 0xff);
	}

	/* Formatting cylinder 0, x to go... */
	UpdateStats(TRUE, 1, 0, numcyls);

	for (curcyl = 0; ret && curcyl < numcyls; curcyl++)
	{
		for (mode = 1; ret && mode >= 0; mode--)
		{
			ULONG pos;

			if (SetSignal(0L, SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C)
			{
				PutStr("\n");
				ioerr = ERROR_BREAK;
				ret = FALSE;
				break;
			}

			/* Reverse (1 - mode) logic here... We fake the display a bit here. If this
			 * looks funny to you, try changing this to be logically correct, and see
			 * what happens with >=99 cylinder format. - Piru
			 */
			UpdateStats(FALSE, 1 - mode, curcyl, numcyls);

			for (pos = 0; pos < tracksize; pos += bufsize)
			{
				ULONG size;

				writebuf[0] = (curcyl == 0 && pos == 0) ? ID_UNREADABLE_DISK : (dostype & 0xffffff00);

				offset = (UQUAD) (startcyl + curcyl) * (UQUAD) tracksize + pos;
				size = tracksize - pos;
				size = size > bufsize ? bufsize : size;

				DEBUG_FMTPART(Printf("%s at offset %llu, %lu bytes\n", (LONG) (mode ? "Format" : "Read"),
				                     (ULONG) (offset >> 32), (ULONG) offset, size));

				if (usetd64)
				{
					IOReq->io_Command    = mode ? CMD_FORMAT64 : CMD_READ64;
					IOReq->io_HighOffset = (ULONG) (offset >> 32);
					IOReq->io_LowOffset  = (ULONG) offset;
				}
				else
				{
					IOReq->io_Command = mode ? TD_FORMAT : CMD_READ;
					IOReq->io_Offset  = offset;
				}
				IOReq->io_Flags  = 0;
				IOReq->io_Length = size;
				IOReq->io_Data   = mode ? writebuf : readbuf;

				err = DoIO((APTR) IOReq);

				if (!err)
				{
					if (mode)
					{
						/* We did some successful writes to disk, remember this for
						 * later CMD_UPDATE and CMD_FLUSH.
						 */
						WritesDone = TRUE;
					}
					else
					{
						/* We actually check if the data we read back is what we wrote.
						 * Original format ignore the read data, and assume it was ok if
						 * the CMD_READ succeeds in the first place. - Piru
						 */
						if (memcmp(readbuf, writebuf, size) != 0)
						{
							err = TDERR_BadSecSum;
						}
					}
				}

				if (err)
				{
#warning "TODO: Add retry..."
					/* Show the real cylinder & mode before throwing out the error... */
					UpdateStats(TRUE, mode, curcyl, numcyls);
					PutStr("\n");
					Printf(GetStr(MSG_BODY_COULD_NOT_FORMAT_CYLINDER), curcyl);
					Printf(": %s\n", (LONG) ConvDevErr(err, &ioerr));
					ret = FALSE;
					break;
				}
			}
		}
	}

	if (ret)
	{
		/* Verifying  cylinder x, 0 to go... */
		UpdateStats(TRUE, 0, curcyl, numcyls);
	}

	FreeMem(memptr, memlen);

	SetIoErr(ioerr);

	return ret;
}


static
void UpdateStats(BOOL printnow, LONG mode, ULONG curcyl, ULONG numcyls)
{
	if (numcyls >= 999)
	{
		if ((curcyl % 25) == 0)
			printnow = TRUE;
	}
	else if (numcyls >= 99)
	{
		if ((curcyl % 10) == 0)
			printnow = TRUE;
	}
	else
	{
		printnow = TRUE;
	}
	
	if (printnow)
	{
		Printf(GetStr(mode ? MSG_FORMATTING_CYLINDER : MSG_VERIFYING_CYLINDER), curcyl, numcyls - curcyl);
		Flush(Output());
	}
}

static
CONST_STRPTR ConvDevErr(LONG err, LONG *ioerr)
{
	static UBYTE errbuf[32];
	CONST_STRPTR mes;

	*ioerr = 0;
	switch (err)
	{
		case IOERR_OPENFAIL:
			mes = "Device/unit failed to open";
			break;
		case IOERR_ABORTED:
			mes = "Request terminated early [after AbortIO()]";
			break;
		case IOERR_NOCMD:
			*ioerr = ERROR_NOT_IMPLEMENTED;
			mes = "Command not supported by device";
			break;
		case IOERR_BADLENGTH:
			mes = "Not a valid length (usually IO_LENGTH)";
			break;
		case IOERR_BADADDRESS:
			mes = "Invalid address (misaligned or bad range)";
			break;
		case IOERR_UNITBUSY:
			mes = "Device opens ok, but requested unit is busy";
			break;
		case IOERR_SELFTEST:
			mes = "Hardware failed self-test";
			break;

		case TDERR_NotSpecified:
			mes = "General catchall";
			break;
		case TDERR_NoSecHdr:
			mes = "Sector not found";
			break;
		case TDERR_BadSecPreamble:
			mes = "Bad sector (BadSecPreamble)";
			break;
		case TDERR_BadSecID:
			mes = "Bad sector (BadSecID)";
			break;
		case TDERR_BadHdrSum:
			mes = "Incorrect header checksum";
			break;
		case TDERR_BadSecSum:
			mes = "Incorrect data checksum";
			break;
		case TDERR_TooFewSecs:
			mes = "Not enough sectors";
			break;
		case TDERR_BadSecHdr:
			mes = "Bad sector (BadSecHdr)";
			break;
		case TDERR_WriteProt:
			*ioerr = ERROR_WRITE_PROTECTED;
			mes = "Drive is write protected";
			break;
		case TDERR_DiskChanged:
			*ioerr = ERROR_NO_DISK;
			mes = "No disk in drive";
			break;
		case TDERR_SeekError:
			*ioerr = ERROR_SEEK_ERROR;
			mes = "Seek error";
			break;
		case TDERR_NoMem:
			*ioerr = ERROR_NO_FREE_STORE;
			mes = "Out of memory";
			break;
		case TDERR_BadUnitNum:
			mes = "Bad unit number";
			break;
		case TDERR_BadDriveType:
			mes = "Bad drive type";
			break;
		case TDERR_DriveInUse:
			*ioerr = ERROR_OBJECT_IN_USE;
			mes = "Drive in use";
			break;
		case TDERR_PostReset:
			mes = "User hit reset; awaiting doom";
			break;

		case HFERR_SelfUnit:
			mes = "Cannot issue SCSI command to self";
			break;
		case HFERR_DMA:
			mes = "DMA error";
			break;
		case HFERR_Phase:
			mes = "Illegal or unexpected SCSI phase";
			break;
		case HFERR_Parity:
			mes = "SCSI parity error";
			break;
		case HFERR_SelTimeout:
			mes = "Select timed out";
			break;
		case HFERR_BadStatus:
			mes = "Status and/or sense error";
			break;
		case HFERR_NoBoard:
			mes = "Open failed for non-existant board";
			break;

		default:
#ifdef __MORPHOS__
		    RawDoFmt("I/O-error number %ld", &err, (void (*)(void)) 0, errbuf);
#else
			sprintf(errbuf, "I/O-error number %ld", err);
#endif
			mes = errbuf;
			break;
	}

	return mes;
}


static
ULONG CreateIcons(CONST_STRPTR Volume)
{
	ULONG Status = FALSE;
	struct Library *IconBase;

	IconBase = OpenLibrary(ICONNAME, 37);
	if (IconBase)
	{
		ULONG MyLength;
		UBYTE *MyName;

		MyLength = strlen(Volume);
		MyName = AllocVec(MyLength + 1 + 9 + 1,
		                  MEMF_PUBLIC | MEMF_CLEAR);
		if (MyName)
		{
			struct DiskObject *MyIcon;

			memcpy(MyName, Volume,  MyLength);
			MyName[MyLength] = ':';

			MyIcon = GetDefDiskObject(WBDISK);
			if (MyIcon)
			{
				memcpy(MyName + MyLength + 1, "Disk", sizeof("Disk"));

				/* Unsnapshot the icon. - Piru */
				MyIcon->do_CurrentX = NO_ICON_POSITION;
				MyIcon->do_CurrentY = NO_ICON_POSITION;

				if (PutDiskObject(MyName, MyIcon))
				{
					Status = TRUE;
				}

				FreeDiskObject(MyIcon);
			}

			if (Status)
			{
				Status = FALSE;

				MyIcon = GetDefDiskObject(WBGARBAGE);
				if (MyIcon)
				{
					BPTR lock;

					memcpy(MyName + MyLength + 1, "Trashcan", sizeof("Trashcan"));

					lock = CreateDir(MyName);
					if (lock)
					{
						UnLock(lock);

						if (PutDiskObject(MyName, MyIcon))
						{
							Status = TRUE;
						}
					}

					FreeDiskObject(MyIcon);
				}
			}

			FreeVec(MyName);
		}

		if (!Status)
		{
			PrintFault(IoErr(), NULL);
		}

		CloseLibrary(IconBase);
	}

	return Status;
}
