/*
 * $Id: networkprefs.c,v 1.7 2013/05/19 10:47:42 cyfm Exp $
 */
#define AROS_ALMOST_COMPATIBLE
#include "global.h"

#define DEBUG 0
#include <aros/debug.h>

BOOL CheckForNetstack(void)
{
BOOL HaveNetStack=FALSE;
struct Process *me = (struct Process *) FindTask(NULL);
APTR oldwinptr = me->pr_WindowPtr;
BPTR MyLock;

	D(bug("%s:\n",__func__));

	me->pr_WindowPtr = (APTR) -1;

	if ((MyLock = Lock("s:network-startup", ACCESS_READ)))
	{
		UnLock(MyLock);
		HaveNetStack = TRUE;
		D(bug("%s: success\n",__func__));
	}
	else
	{
		D(bug("%s: failed \n",__func__));
	}
	me->pr_WindowPtr = oldwinptr;
	return(HaveNetStack);
}

/*********************************************************************************************/

static void WirelessHandler(void)
{
	struct Process *me = (struct Process *) FindTask(NULL);
	#if 0
	APTR oldwinptr = me->pr_WindowPtr;
	#endif
	struct MsgPort notifyport;
	struct NotifyRequest nr;

	D(bug("%s:\n",__func__));

	me->pr_WindowPtr = (APTR) -1;

	notifyport.mp_Node.ln_Type = NT_MSGPORT;
	notifyport.mp_SigBit  = SIGBREAKB_CTRL_F;
	notifyport.mp_SigTask = &me->pr_Task;
	notifyport.mp_Flags   = PA_SIGNAL;
	NEWLIST(&notifyport.mp_MsgList);

	memset(&nr, 0, sizeof(nr));
	nr.nr_Name  = "ENV:sys/WLANMNGR";
	nr.nr_Flags = NRF_SEND_MESSAGE | NRF_NOTIFY_INITIAL;
	nr.nr_stuff.nr_Msg.nr_Port = &notifyport;
	if (!StartNotify(&nr))
	{
		D(bug("%s: StartNotify failed\n",__func__));
		return;
	}

	for(;;)
	{
		struct NotifyMessage *nm;
		BOOL changed = FALSE;
		BPTR MyLock;
		BPTR ifh, ofh;

		WaitPort(&notifyport);

		while ((nm = (struct NotifyMessage *) GetMsg(&notifyport)))
		{
			ReplyMsg(&nm->nm_ExecMessage);
			changed = TRUE;
		}
		if (!changed)
		{
			/* Spurious signal? - do nothing */
			continue;
		}

		MyLock = Lock("ENV:sys/WLANMNGR", ACCESS_READ);
		if (!MyLock)
		{
			/* File was deleted? - do nothing */
			continue;
		}

		UnLock(MyLock);

		if(FindPort("WLANMNGR_GLOBAL") != NULL)
		{
			/* Wireless manager is already running - do nothing */
			continue;
		}

		D(bug("%s: WLANMNGR env found, WirelessManager will be started async\n",__func__));

		ifh = Open("NIL:", MODE_OLDFILE);
		ofh = Open("NIL:", MODE_NEWFILE);

		if (SystemTags("MOSSYS:Net/WirelessManager",
		               SYS_Asynch, TRUE,
		               SYS_Input, ifh,
		               SYS_Output, ofh,
		               TAG_END) != -1)
		{
			/* successful launch, handles are closed elsewhere */
			ifh = 0;
			ofh = 0;
		}

		if (ofh)
		{
			Close(ofh);
		}

		if (ifh)
		{
			Close(ifh);
		}
	}

#if 0 /* never reached */
	EndNotify(&nr);

	me->pr_WindowPtr = oldwinptr;
#endif
}

/*********************************************************************************************/

#define	WLANHANDLER_NAME	"WLanNetworkHandler"

void LoadNetworkPrefs(char *filename)
{
	D(bug("%s: <%s>\n",__func__,filename));

	if (CheckForNetstack())
	{
		BPTR ifh, ofh;
		struct Process	*MyProcess;

		D(bug("%s: Launch WirelessHandler\n",__func__));

		MyProcess = CreateNewProcTags(NP_Entry,(ULONG) &WirelessHandler,
		                              NP_Name,(ULONG) WLANHANDLER_NAME,
		                              NP_Priority,0,
		                              NP_CodeType,MACHINE_PPC,
		                              NP_ConsoleTask, NULL,
		                              NP_CurrentDir, 0,
		                              NP_CopyVars, FALSE,
		                              NP_WindowPtr, NULL,
		                              NP_HomeDir, 0,
		                              TAG_END);


		D(bug("%s: <s:network-startup> exists, use it\n",__func__));

		ifh = Open("NIL:", MODE_OLDFILE);
		ofh = Open("NIL:", MODE_NEWFILE);
		if (ifh && ofh)
		{
			#if 1
			if (SystemTags("execute s:network-startup",
			               SYS_Asynch, TRUE,
			               SYS_Input, ifh,
			               SYS_Output, ofh,
			               TAG_END) != -1)
			{
				ifh = 0;
				ofh = 0;
			}
			#else
			SystemTags("execute s:network-startup",
			           SYS_Asynch, TRUE,
			           SYS_Input, NULL,
			           SYS_Output, NULL,
			           TAG_END);
			#endif

			D(bug("%s: SystemTags() returned\n",__func__));
		}

		if (ofh)
		{
			Close(ofh);
		}

		if (ifh)
		{
			Close(ifh);
		}
	}
	else
	{
		D(bug("%s: NetStack disabled\n",__func__));
	}
}

/*********************************************************************************************/

void StartNetwork(void)
{
struct Process *me = (struct Process *) FindTask(NULL);
APTR oldwinptr = me->pr_WindowPtr;
BPTR MyLock;

	D(bug("%s:\n",__func__));

	me->pr_WindowPtr = (APTR) -1;

	if ((MyLock = Lock("env:sys/interface.prefs", ACCESS_READ)))
	{
		UnLock(MyLock);
		D(bug("%s: config file found, network will be started async\n",__func__));
	}
	else
	{
		D(bug("%s: no config file found, start network direct\n",__func__));
		LoadNetworkPrefs(NULL);
	}
	me->pr_WindowPtr = oldwinptr;
}
